from ._objects import Objects


class ObjectSet:
    """A set of objects.Objects instances.

    This class allows you to either refer to an object by name or
    iterate over all available objects.
    """

    def __init__(self, can_overwrite=False):
        """Initialize the object set

        can_overwrite - True to allow overwriting of a new copy of objects
                        over an old one of the same name (for debugging)
        """
        self.__can_overwrite = can_overwrite
        self.__types_and_instances = {"objects": {}}

    @property
    def __objects_by_name(self):
        return self.__types_and_instances["objects"]

    def add_objects(self, objects, name):
        assert isinstance(
            objects, Objects
        ), "objects must be an instance of CellProfiler.Objects"
        assert (
            name not in list(self.__objects_by_name.keys())
        ) or self.__can_overwrite, (
            "The object, {}, is already in the object set".format(name)
        )
        self.__objects_by_name[name] = objects

    def get_object_names(self):
        """Return the names of all of the objects
        """
        return list(self.__objects_by_name.keys())

    object_names = property(get_object_names)

    def get_objects(self, name):
        """Return the objects instance with the given name
        """
        return self.__objects_by_name[name]

    @property
    def all_objects(self):
        """Return a list of name / objects tuples
        """
        return list(self.__objects_by_name.items())

    def get_types(self):
        """Get then names of types of per-image set "things"

        The object set can store arbitrary types of things other than objects,
        for instance ImageJ data tables. This function returns the thing types
        defined in the object set at this stage of the pipeline.
        """
        return list(self.__types_and_instances.keys())

    def add_type_instance(self, type_name, instance_name, instance):
        """Add a named instance of a type

        A thing of a given type can be stored in the object set so that
        it can be retrieved by name later in the pipeline. This function adds
        an instance of a type to the object set.

        type_name - the name of the instance's type
        instance_name - the name of the instance
        instance - the instance itself
        """
        if type_name not in self.__types_and_instances:
            self.__types_and_instances[type_name] = {}
        self.__types_and_instances[type_name][instance_name] = instance

    def get_type_instance(self, type_name, instance_name):
        """Get an named instance of a type

        type_name - the name of the type of instance
        instance_name - the name of the instance to retrieve
        """
        if (
            type_name not in self.__types_and_instance
            or instance_name not in self.__types_and_instances[type_name]
        ):
            return None
        return self.__types_and_instances[type_name][instance_name]



"""
java.py - CellProfiler-specific JVM utilities
"""

import logging
import os
import scyjava

import cellprofiler_core.preferences


LOGGER = logging.getLogger(__name__)

def start_java():
    """Start CellProfiler's JVM via Javabridge

    JVM parameters are harvested from preferences and
    the environment variables:

    CP_JDWP_PORT - port # for debugging Java within the JVM
    cpprefs.get_awt_headless() - controls java.awt.headless to prevent
        awt from being invoked
    """
    if scyjava.jvm_started():
        return

    # Add Bio-Formats Java dependency.
    scyjava.config.endpoints.append("ome:formats-gpl")
    scyjava.config.endpoints.append("org.scijava:scijava-config")

    LOGGER.info("Initializing Java Virtual Machine")
    args = [
        "-Dloci.bioformats.loaded=true",
    ]

    awt_headless = cellprofiler_core.preferences.get_awt_headless()
    if awt_headless:
        LOGGER.debug("JVM will be started with AWT in headless mode")
        args.append("-Djava.awt.headless=true")

    if "CP_JDWP_PORT" in os.environ:
        args.append(
            (
                "-agentlib:jdwp=transport=dt_socket,address=127.0.0.1:%s"
                ",server=y,suspend=n"
            )
            % os.environ["CP_JDWP_PORT"]
        )
    scyjava.start_jvm(options=args)
    #
    # Enable Bio-Formats directory cacheing
    #
    Location = scyjava.jimport("loci.common.Location")
    Location.cacheDirectoryListings(True)
    LOGGER.debug("Enabled Bio-formats directory cacheing")


def stop_java():
    LOGGER.info("Shutting down Java Virtual Machine")
    scyjava.shutdown_jvm()

def jimport(package):
    start_java()
    return scyjava.jimport(package)



import numpy
from packaging.version import Version

from cellprofiler_core import __version__ as core_version

IMAGE_NUMBER = "ImageNumber"
GROUP_NUMBER = "Group_Number"
GROUP_INDEX = "Group_Index"
GROUP_LENGTH = "Group_Length"
CURRENT = "Current"
NUMBER_OF_IMAGE_SETS = "NumberOfImageSets"
NUMBER_OF_MODULES = "NumberOfModules"
SET_BEING_ANALYZED = "SetBeingAnalyzed"
SAVE_OUTPUT_HOW_OFTEN = "SaveOutputHowOften"
TIME_STARTED = "TimeStarted"
STARTING_IMAGE_SET = "StartingImageSet"
STARTUP_DIRECTORY = "StartupDirectory"
DEFAULT_MODULE_DIRECTORY = "DefaultModuleDirectory"
DEFAULT_IMAGE_DIRECTORY = "DefaultImageDirectory"
DEFAULT_OUTPUT_DIRECTORY = "DefaultOutputDirectory"
IMAGE_TOOLS_FILENAMES = "ImageToolsFilenames"
IMAGE_TOOL_HELP = "ImageToolHelp"
PREFERENCES = "Preferences"
PIXEL_SIZE = "PixelSize"
SKIP_ERRORS = "SkipErrors"
INTENSITY_COLOR_MAP = "IntensityColorMap"
LABEL_COLOR_MAP = "LabelColorMap"
STRIP_PIPELINE = "StripPipeline"
DISPLAY_MODE_VALUE = "DisplayModeValue"
DISPLAY_WINDOWS = "DisplayWindows"
FONT_SIZE = "FontSize"
IMAGES = "Images"
MEASUREMENTS = "Measurements"
PIPELINE = "Pipeline"
SETTINGS = "Settings"
VARIABLE_VALUES = "VariableValues"
VARIABLE_INFO_TYPES = "VariableInfoTypes"
MODULE_NAMES = "ModuleNames"
NUMBERS_OF_VARIABLES = "NumbersOfVariables"
VARIABLE_REVISION_NUMBERS = "VariableRevisionNumbers"
MODULE_REVISION_NUMBERS = "ModuleRevisionNumbers"
MODULE_NOTES = "ModuleNotes"
CURRENT_MODULE_NUMBER = "CurrentModuleNumber"
SHOW_WINDOW = "ShowFrame"
BATCH_STATE = "BatchState"
EXIT_STATUS = "Exit_Status"
SETTINGS_DTYPE = numpy.dtype(
    [
        (VARIABLE_VALUES, "|O4"),
        (VARIABLE_INFO_TYPES, "|O4"),
        (MODULE_NAMES, "|O4"),
        (NUMBERS_OF_VARIABLES, "|O4"),
        (PIXEL_SIZE, "|O4"),
        (VARIABLE_REVISION_NUMBERS, "|O4"),
        (MODULE_REVISION_NUMBERS, "|O4"),
        (MODULE_NOTES, "|O4"),
        (SHOW_WINDOW, "|O4"),
        (BATCH_STATE, "|O4"),
    ]
)


def make_cell_struct_dtype(fields):
    """Makes the dtype of a struct composed of cells

    fields - the names of the fields in the struct
    """
    return numpy.dtype([(str(x), "|O4") for x in fields])


CURRENT_DTYPE = make_cell_struct_dtype(
    [
        NUMBER_OF_IMAGE_SETS,
        SET_BEING_ANALYZED,
        NUMBER_OF_MODULES,
        SAVE_OUTPUT_HOW_OFTEN,
        TIME_STARTED,
        STARTING_IMAGE_SET,
        STARTUP_DIRECTORY,
        DEFAULT_OUTPUT_DIRECTORY,
        DEFAULT_IMAGE_DIRECTORY,
        IMAGE_TOOLS_FILENAMES,
        IMAGE_TOOL_HELP,
    ]
)
PREFERENCES_DTYPE = make_cell_struct_dtype(
    [
        PIXEL_SIZE,
        DEFAULT_MODULE_DIRECTORY,
        DEFAULT_OUTPUT_DIRECTORY,
        DEFAULT_IMAGE_DIRECTORY,
        INTENSITY_COLOR_MAP,
        LABEL_COLOR_MAP,
        STRIP_PIPELINE,
        SKIP_ERRORS,
        DISPLAY_MODE_VALUE,
        FONT_SIZE,
        DISPLAY_WINDOWS,
    ]
)
NATIVE_VERSION = Version(core_version).major
IMAGE_PLANE_DESCRIPTOR_VERSION = 1
H_SVN_REVISION = "SVNRevision"
H_DATE_REVISION = "DateRevision"
H_GIT_HASH = "GitHash"
H_PLANE_COUNT = "PlaneCount"
H_URL = "URL"
H_SERIES = "Series"
H_INDEX = "Index"
H_CHANNEL = "Channel"
H_MODULE_COUNT = "ModuleCount"
H_HAS_IMAGE_PLANE_DETAILS = "HasImagePlaneDetails"
H_MESSAGE_FOR_USER = "MessageForUser"
COOKIE = "CellProfiler Pipeline: http://www.nucleus.org"
SAD_PROOFPOINT_COOKIE = r"CellProfiler Pipeline: https?://\S+.proofpoint.com.+http-3A__www.cellprofiler\.org"
HDF5_HEADER = (
    chr(137) + chr(72) + chr(68) + chr(70) + chr(13) + chr(10) + chr(26) + chr(10)
)
C_PIPELINE = "Pipeline"
C_CELLPROFILER = "CellProfiler"
F_PIPELINE = "Pipeline"
F_USER_PIPELINE = "UserPipeline"
M_PIPELINE = "_".join((C_PIPELINE, F_PIPELINE))
M_USER_PIPELINE = "_".join((C_PIPELINE, F_USER_PIPELINE))
F_VERSION = "Version"
M_VERSION = "_".join((C_CELLPROFILER, F_VERSION))
C_RUN = "Run"
C_MODIFICATION = "Modification"
F_TIMESTAMP = "Timestamp"
M_TIMESTAMP = "_".join((C_RUN, F_TIMESTAMP))
M_MODIFICATION_TIMESTAMP = "_".join((C_MODIFICATION, F_TIMESTAMP))
M_DEFAULT_INPUT_FOLDER = "Default_InputFolder"
M_DEFAULT_OUTPUT_FOLDER = "Default_OutputFolder"
RF_STATE_PREQUOTE = 0
RF_STATE_FIELD = 1
RF_STATE_BACKSLASH_ESCAPE = 2
RF_STATE_SEPARATOR = 3
DIRECTION_UP = "up"
DIRECTION_DOWN = "down"



from ...utilities.zmq.communicable.request import AnalysisRequest


class SharedDictionary(AnalysisRequest):
    def __init__(self, analysis_id, module_num=-1):
        AnalysisRequest.__init__(self, analysis_id, module_num=module_num)



import functools

from ._module import Module
from ..constants.measurement import C_CHILDREN
from ..constants.measurement import C_COUNT
from ..constants.measurement import C_LOCATION
from ..constants.measurement import C_NUMBER
from ..constants.measurement import C_PARENT
from ..constants.measurement import FTR_CENTER_X
from ..constants.measurement import FTR_CENTER_Y
from ..constants.measurement import FTR_OBJECT_NUMBER
from ..constants.measurement import IMAGE


class Identify(Module):
    @staticmethod
    def get_object_categories(pipeline, object_name, object_dictionary):
        """Get categories related to creating new children

        pipeline - the pipeline being run (not used)
        object_name - the base object of the measurement: "Image" or an object
        object_dictionary - a dictionary where each key is the name of
                            an object created by this module and each
                            value is a list of names of parents.
        """
        if object_name == IMAGE:
            return [C_COUNT]
        result = []
        if object_name in object_dictionary:
            result += [
                C_LOCATION,
                C_NUMBER,
            ]
            if len(object_dictionary[object_name]) > 0:
                result += [C_PARENT]
        if object_name in functools.reduce(
            lambda x, y: x + y, list(object_dictionary.values())
        ):
            result += [C_CHILDREN]
        return result

    @staticmethod
    def get_object_measurements(pipleline, object_name, category, object_dictionary):
        """Get measurements related to creating new children

        pipeline - the pipeline being run (not used)
        object_name - the base object of the measurement: "Image" or an object
        object_dictionary - a dictionary where each key is the name of
                            an object created by this module and each
                            value is a list of names of parents.
        """
        if object_name == IMAGE and category == C_COUNT:
            return list(object_dictionary.keys())

        if object_name in object_dictionary:
            if category == C_LOCATION:
                return [
                    FTR_CENTER_X,
                    FTR_CENTER_Y,
                ]
            elif category == C_NUMBER:
                return [FTR_OBJECT_NUMBER]
            elif category == C_PARENT:
                return list(object_dictionary[object_name])
        if category == C_CHILDREN:
            result = []
            for child_object_name in list(object_dictionary.keys()):
                if object_name in object_dictionary[child_object_name]:
                    result += ["%s_Count" % child_object_name]
            return result
        return []



from ._setting import Setting


class HiddenCount(Setting):
    """A setting meant only for saving an item count

    The HiddenCount setting should never be in the visible settings.
    It should be tied to a sequence variable which gives the number of
    items which is the value for this variable.
    """

    def __init__(self, sequence, text="Hidden"):
        super(HiddenCount, self).__init__(text, str(len(sequence)))
        self.__sequence = sequence

    def set_value(self, value):
        if not value.isdigit():
            raise ValueError("The value must be an integer")
        count = int(value)
        if count == len(self.__sequence):
            # The value was "inadvertantly" set, but is correct
            return
        raise NotImplementedError(
            "The count should be inferred, not set  - actual: %d, set: %d"
            % (len(self.__sequence), count)
        )

    def get_value(self):
        return len(self.__sequence)

    def set_sequence(self, sequence):
        """Set the sequence used to maintain the count"""
        self.__sequence = sequence

    def __str__(self):
        return str(len(self.__sequence))

    def get_unicode_value(self):
        return str(len(self.__sequence))



from enum import Enum
from .._setting import Setting
from .._validation_error import ValidationError


class Choice(Setting):
    """A setting that displays a drop-down set of choices

    """

    def __init__(
        self, text, choices, value=None, tooltips=None, choices_fn=None, *args, **kwargs
    ):
        """Initializer
        module - the module containing the setting
        text - the explanatory text for the setting
        choices - a sequence of string choices to be displayed in the drop-down
                  if an enum is provided, member values are used as choices
        value - the default choice or None to choose the first of the choices.
        tooltips - a dictionary of choice to tooltip
        choices_fn - a function that, if present, supplies the choices. The
                     function should have the signature, fn(pipeline).
        """
        if isinstance(choices, type(Enum)):
            self.__enum = choices
            choices = [member.value for member in self.__enum]

        super(Choice, self).__init__(text, value or choices[0], *args, **kwargs)
        self.__choices = choices
        self.__tooltips = tooltips
        self.__choices_fn = choices_fn

    def __internal_get_choices(self):
        """The sequence of strings that define the choices to be displayed"""
        return self.get_choices()

    def get_choices(self):
        """The sequence of strings that define the choices to be displayed"""
        return self.__choices

    choices = property(__internal_get_choices)

    def __internal_get_enum_member(self):
        return self.get_enum_member()
    
    def get_enum_member(self):
        """The enum value for the current setting"""
        if self.__enum:
            return self.__enum(self.value)
        else:
            raise ValueError("Choice setting is not an enum")
    
    enum_member = property(__internal_get_enum_member)

    def get_tooltips(self):
        """The tooltip strings for each choice"""
        return self.__tooltips

    tooltips = property(get_tooltips)

    @property
    def has_tooltips(self):
        """Return true if the choice has tooltips installed"""
        return self.__tooltips is not None

    def test_valid(self, pipeline):
        """Check to make sure that the value is among the choices"""
        if self.__choices_fn is not None:
            self.__choices = self.__choices_fn(pipeline)
        if self.value not in self.choices:
            raise ValidationError(
                "%s is not one of %s" % (self.value, ",".join(self.choices)), self
            )



import json

from ._setting import Setting


class FileCollectionDisplay(Setting):
    """A setting to be used to display directories and their files

    The FileCollectionDisplay shows directory trees with mechanisms to
    communicate directory additions and deletions to its parent module.

    The central data structure is the dictionary, "self.file_tree". The keys
    for the top-level of the dictionary are the directories managed by the
    setting. If a key represents a directory, its value is another directory.
    If a key represents a file, its value is either True (the file is included
    in the collection) or False (the file is filtered out of the collection).

    Directory dictionaries can be filtered: this is done by setting the
    special key, "None" to either True or False.

    The FileCollectionDisplay manages the tree and it should be treated as
    read-only by callers. Callers can request that nodes be added, removed,
    filtered or not filtered by calling the appropriate notification function
    with a nested collection of two-tuples and strings (modpaths). Two-tuples
    represent directories whose subdirectories or files are being operated on.
    Strings represent directories or files that are being operated on. The first
    element of the two-tuple is the directory name and the second is a
    sub-collection of two-tuples. For instance, to operate on foo/bar, send:

    ("foo", ("bar", ))

    The FileCollectionDisplay communicates events on individual files or
    directories by specifying a path as a collection of path parts. These
    can be any sort of object and it is the caller's job to maintain the
    display names of each of them and their node categories (used for
    icon display).
    """

    ADD = "ADD"
    REMOVE = "REMOVE"
    METADATA = "METADATA"
    NODE_DIRECTORY = "directory"
    NODE_COMPOSITE_IMAGE = "compositeimage"
    NODE_COLOR_IMAGE = "colorimage"
    NODE_MONOCHROME_IMAGE = "monochromeimage"
    NODE_IMAGE_PLANE = "imageplane"
    NODE_MOVIE = "movie"
    NODE_FILE = "file"
    NODE_CSV = "csv"
    BKGND_PAUSE = "pause"
    BKGND_RESUME = "resume"
    BKGND_STOP = "stop"
    BKGND_GET_STATE = "getstate"

    class DeleteMenuItem(object):
        """A placeholder in the context menu for the delete command

        The DeleteMenuItem can be placed in the context menu returned
        by fn_get_path_info so that the user can delete the selected items
        from the context menu.

        text - the text to display in the context menu
        """

        def __init__(self, text):
            self.text = text

    def __init__(
        self,
        text,
        value,
        fn_on_drop,
        fn_on_remove,
        fn_get_path_info,
        fn_on_menu_command,
        fn_on_bkgnd_control,
        hide_text="Hide filtered files",
        **kwargs,
    ):
        """Constructor

        text - the label to the left of the setting

        value - the value for the control. This is a serialization of
                the appearance (for instance, whether to show or hide
                filtered files).

        fn_on_drop - called when files are dropped. The signature is
                     fn_on_drop(pathnames, check_for_directories) The first
                     argument is a list of pathnames of the dropped files.
                     The second argument is True if the user has performed
                     a file name drop which might include directories and
                     False if the user has dropped text file names.

        fn_on_remove - called when the UI requests that files be removed. Has
                       one argument which is a collection of paths to remove.

        fn_get_path_info - called when the UI needs to know the display name,
                     icon type, context menu and tool tip for an item. These
                     are returned in a four-tuple by the callee, e.g:
                     [ "image.tif", NODE_MONOCHROME_IMAGE,
                       "image of well A01 on plate P-12345",
                       ( "Show image", "Show metadata", "Delete image")]

        fn_on_menu_command - called when the user selects a context menu
                     command. The argument is the text from the context menu or
                     None if the default command.

        fn_on_bkgnd_control - called when the UI wants to stop, pause or resume
                     all background processing. BKGND_PAUSE asks for the
                     caller to pause processing, BKGND_RESUME asks for the
                     caller to resume, BKGND_STOP asks for processing to be
                     aborted, BKGND_GET_STATE asks for the caller to
                     return its current state = BKGND_PAUSE if it is paused,
                     BKGND_RESUME if it is running or BKGND_STOP if it is
                     idle.

        hide_text - the text displayed next to the hide checkbox.
        """
        super(self.__class__, self).__init__(text, value, **kwargs)
        self.fn_on_drop = fn_on_drop
        self.fn_on_remove = fn_on_remove
        self.fn_get_path_info = fn_get_path_info
        self.fn_on_menu_command = fn_on_menu_command
        self.fn_on_bkgnd_control = fn_on_bkgnd_control
        self.hide_text = hide_text
        self.fn_update = None
        self.file_tree = {}
        self.properties = {self.SHOW_FILTERED: True}
        try:
            properties = json.loads(value)
            if isinstance(properties, dict):
                self.properties.update(properties)
        except:
            pass

    SHOW_FILTERED = "ShowFiltered"

    def update_value(self):
        """Update the setting value after changing a property"""
        self.value_text = json.dumps(self.properties)

    def update_ui(self, cmd=None, mods=None):
        if self.fn_update is not None:
            self.fn_update(cmd, mods)

    def set_update_function(self, fn_update=None):
        """Set the function that will be called when the file_tree is updated"""
        self.fn_update = fn_update

    def initialize_tree(self, mods):
        """Remove all nodes in the file tree"""
        self.file_tree = {}
        self.add_subtree(mods, self.file_tree)

    def add(self, mods):
        """Add nodes to the file tree

        mods - modification structure. See class documentation for its form.
        """
        self.add_subtree(mods, self.file_tree)
        self.update_ui(self.ADD, mods)

    def modify(self, mods):
        """Indicate a minor modification such as metadtaa change

        mods - modification structure. See class documentation for its form.
        """
        self.update_ui(self.METADATA, mods)

    @classmethod
    def is_leaf(cls, mod):
        """True if the modification structure is the leaf of a tree

        The leaves are either strings representing the last part of a path
        or 3-tuples representing image planes within an image file. Branches
        are two-tuples composed of a path part and more branches / leaves
        """
        return len(mod) != 2 or not isinstance(mod[0], str)

    def node_count(self, file_tree=None):
        """Count the # of nodes (leaves + directories) in the tree"""
        if file_tree is None:
            file_tree = self.file_tree
        count = 0
        for key in list(file_tree.keys()):
            if key is None:
                pass
            elif isinstance(file_tree[key], dict):
                count += 1 + self.node_count(file_tree[key])
            else:
                count += 1
        return count

    def get_tree_modpaths(self, path):
        """Create a modpath containing the selected node and all children

        root - list of paths to the selected node

        returns a modpath (two-tuples where the first is the key and the second
        is a list of sub-modpaths)
        """
        tree = self.file_tree
        root_modlist = sub_modlist = []
        while len(path) > 1:
            next_sub_modlist = []
            sub_modlist.append((path[0], next_sub_modlist))
            tree = tree[path[0]]
            path = path[1:]
            sub_modlist = next_sub_modlist
        if isinstance(tree[path[0]], dict):
            sub_modlist.append((path[0], self.get_all_modpaths(tree[path[0]])))
        else:
            sub_modlist.append(path[0])
        return root_modlist[0]

    def get_all_modpaths(self, tree):
        """Get all sub-modpaths from the branches of the given tree"""
        result = []
        for key in list(tree.keys()):
            if key is None:
                continue
            elif not isinstance(tree[key], dict):
                result.append(key)
            else:
                result.append((key, self.get_all_modpaths(tree[key])))
        return result

    def add_subtree(self, mods, tree):
        for mod in mods:
            if self.is_leaf(mod):
                if mod not in tree:
                    tree[mod] = True
            else:
                if mod[0] in tree and isinstance(tree[mod[0]], dict):
                    subtree = tree[mod[0]]
                else:
                    subtree = tree[mod[0]] = {}
                subtree[None] = True
                self.add_subtree(mod[1], subtree)

    def on_remove(self, mods):
        """Called when the UI wants to remove nodes

        mods - a modlist of nodes to remove
        """
        self.fn_on_remove(mods)

    def remove(self, mods):
        """Remove nodes from the file tree

        mods - modification structure. See class documentation for its form.
        """
        for mod in mods:
            self.remove_subtree(mod, self.file_tree)
        self.update_ui(self.REMOVE, mods)

    def remove_subtree(self, mod, tree):
        if not (isinstance(mod, tuple) and len(mod) == 2):
            if mod in tree:
                subtree = tree[mod]
                if isinstance(subtree, dict):
                    #
                    # Remove whole tree
                    #
                    for key in list(subtree.keys()):
                        if key is None:
                            continue
                        if isinstance(subtree[key], dict):
                            self.remove_subtree(key, subtree)
                del tree[mod]
        elif mod[0] in tree:
            root_mod = mod[0]
            subtree = tree[root_mod]
            if isinstance(subtree, dict):
                for submod in mod[1]:
                    self.remove_subtree(submod, subtree)
                #
                # Delete the subtree if the subtree is emptied
                #
                if len(subtree) == 0 or (len(subtree) == 1 and None in subtree):
                    del tree[root_mod]
            else:
                del tree[root_mod]

    def mark(self, mods, keep):
        """Mark tree nodes as filtered in or out

        mods - modification structure. See class documentation for its form.

        keep - true to mark a node as in the set, false to filter it out.
        """
        self.mark_subtree(mods, keep, self.file_tree)
        self.update_ui()

    def mark_subtree(self, mods, keep, tree):
        for mod in mods:
            if self.is_leaf(mod):
                if mod in tree:
                    if isinstance(tree[mod], dict):
                        tree[mod][None] = keep
                    else:
                        tree[mod] = keep
            else:
                if mod[0] in tree:
                    self.mark_subtree(mod[1], keep, tree[mod[0]])
        kept = [
            tree[k][None] if isinstance(tree[k], dict) else tree[k]
            for k in list(tree.keys())
            if k is not None
        ]
        tree[None] = any(kept)

    def get_node_info(self, path):
        """Get the display name, node type and tool tip for a node

        path - path to the image plane as a list of nodes

        returns a tuple of display name, node type and tool tip
        """
        display_name, node_type, tool_tip, menu = self.fn_get_path_info(path)
        return display_name, node_type, tool_tip

    def get_context_menu(self, path):
        """Get the context menu associated with a path

        path - path to the image plane

        returns a list of context menu items.
        """
        display_name, node_type, tool_tip, menu = self.fn_get_path_info(path)
        return menu

    def get_show_filtered(self):
        return self.properties[self.SHOW_FILTERED]

    def set_show_filtered(self, show_state):
        """Mark that we should show filtered files in the user interface

        show_state - true to show files / false to hide them
        """
        self.properties[self.SHOW_FILTERED] = show_state
        self.update_value()
        self.update_ui()

    show_filtered = property(get_show_filtered, set_show_filtered)



import logging

def set_log_level(log_level, subprocess=False):
    """Set the logging package's log level based on command-line options"""
    try:
        if log_level.isdigit():
            logging.root.setLevel(int(log_level))
        else: # str, e.g. str(logging.INFO) -> '20' OR 'INFO'
            logging.root.setLevel(log_level)

        if len(logging.root.handlers) == 0:
            stream_handler = logging.StreamHandler()
            if subprocess:
                fmt = logging.Formatter("%(process)d|%(levelno)s|%(name)s::%(funcName)s: %(message)s")
            else:
                fmt = logging.Formatter("[CP - %(levelname)s] %(name)s::%(funcName)s: %(message)s")
            stream_handler.setFormatter(fmt)
            logging.root.addHandler(stream_handler)

        # silence matplotlib debug messages, they're super annoying
        logging.getLogger('matplotlib').setLevel(logging.WARNING)
    except ValueError as e:
        logging.config.fileConfig(log_level)



import logging


LOGGER = logging.getLogger(__name__)

class AbstractImage:
    """Represents an image provider that returns images
    """

    def provide_image(self, image_set):
        """Return the image that is associated with the image set
        """
        raise NotImplementedError("Please implement ProvideImage for your class")

    def __get_name(self):
        """Call the abstract function, "get_name"
        """
        return self.get_name()

    def get_name(self):
        """The user-visible name for the image
        """
        raise NotImplementedError("Please implement get_name for your class")

    def release_memory(self):
        """Release whatever memory is associated with the image"""
        LOGGER.warning(
            "Warning: no memory release function implemented for %s image",
            self.get_name(),
        )

    name = property(__get_name)



from ._setting import Setting
from ._validation_error import ValidationError


class Coordinates(Setting):
    """A setting representing X and Y coordinates on an image
    """

    def __init__(self, text, value=(0, 0), *args, **kwargs):
        """Initialize an integer range
        text  - helpful text to be displayed to the user
        value - initial default value, a two-tuple as x and y
        """
        super(Coordinates, self).__init__(text, "%d,%d" % value, *args, **kwargs)

    def set_value(self, value):
        """Convert integer tuples to string
        """
        try:
            if len(value) == 2:
                super(Coordinates, self).set_value("%d,%d" % (value[0], value[1]))
                return
        except:
            pass
        super(Coordinates, self).set_value(value)

    def get_value(self):
        """Convert the underlying string to a two-tuple"""
        return self.get_x(), self.get_y()

    def get_x_text(self):
        """Get the x coordinate as text"""
        return self.get_value_text().split(",")[0]

    def get_x(self):
        """The x coordinate"""
        return int(self.get_x_text())

    x = property(get_x)

    def get_y_text(self):
        vv = self.get_value_text().split(",")
        if len(vv) < 2:
            return ""
        return vv[1]

    def get_y(self):
        """The y coordinate"""
        return int(self.get_y_text())

    y = property(get_y)

    def test_valid(self, pipeline):
        values = self.value_text.split(",")
        if len(values) < 2:
            raise ValidationError("X and Y values must be separated by a comma", self)
        if len(values) > 2:
            raise ValidationError("Only two values allowed", self)
        for value in values:
            if not value.isdigit():
                raise ValidationError("%s is not an integer" % value, self)



from ._event import Event


class ModuleMoved(Event):
    """A module moved up or down

    """

    def __init__(self, module_num, direction, is_image_set_modification):
        super(ModuleMoved, self).__init__(
            is_pipeline_modification=True,
            is_image_set_modification=is_image_set_modification,
        )
        self.module_num = module_num
        self.direction = direction

    def event_type(self):
        return "Module moved"



from ...utilities.zmq.communicable.request import AnalysisRequest


class DebugWaiting(AnalysisRequest):
    """Communicate the debug port to the server and wait for server OK to attach"""

    def __init__(self, analysis_id, port):
        AnalysisRequest.__init__(self, analysis_id=analysis_id, port=port)



from ._abstract_image import AbstractImage


class CallbackImage(AbstractImage):
    """An image provider proxy that calls the indicated callback functions (presumably in your module) to implement the methods
    """

    def __init__(self, name, image_provider_fn):
        """Constructor
        name              - name returned by the Name method
        image_provider_fn - function called during ProvideImage with the arguments, image_set and the CallbackImageProvider instance
        """

        self.__name = name
        self.__image_provider_fn = image_provider_fn

    def provide_image(self, image_set):
        return self.__image_provider_fn(image_set, self)

    def get_name(self):
        return self.__name



import math
import numpy

from ..utilities.image import crop_image


class Image:
    """
    image - a Numpy array

    mask - a binary image indicating the points of interest in the image. The
    mask is the same size as the child image.

    crop_mask - the binary image used to crop the parent image to the
    dimensions of the child (this) image. The crop_mask is the same size as
    the parent image.

    parent_image - for derived images, the parent that was used to create this
    image. This image may inherit attributes from the parent image, such as
    the masks used to create the parent

    masking_objects - the labels matrix from these objects is used to mask and
    crop the parent image to make this image. The labels are available as
    mask_labels and crop_labels.

    convert - true to try to coerce whatever dtype passed (other than bool or
    float) to a scaled image.

    path_name - the path name to the file holding the image or None for a
    derived image

    file_name - the file name of the file holding the image or None for a
    derived image

    scale - the scaling suggested by the initial image format (e.g., 4095 for
    a 12-bit a/d converter).

    Resolution of mask and cropping_mask properties:

    The Image class looks for the mask and cropping_mask in the following
    places:

    * self: if set using the properties or specified in the initializer

    * masking_objects: if set using the masking_object property or specified
    in the initializer. The crop_mask and mask are composed of all of the
    labeled points.

    * parent_image: if set using the initializer. The child image inherits the
    mask and cropping mask of the parent.

    Otherwise, the image has no mask or cropping mask and all pixels are
    significant.
    """

    def __init__(
        self,
        image=None,
        mask=None,
        crop_mask=None,
        parent_image=None,
        masking_objects=None,
        convert=True,
        path_name=None,
        file_name=None,
        scale=None,
        dimensions=2,
        spacing=None,
        channelstack=False
    ):
        self.__image = None

        self.__mask = None

        self.__has_mask = False

        self.parent_image = parent_image

        self.__crop_mask = None

        self.__has_crop_mask = False

        if crop_mask is not None:
            self.crop_mask = crop_mask

        self.masking_objects = masking_objects

        self.__scale = scale

        if image is not None:
            self.set_image(image, convert)

        if mask is not None:
            self.mask = mask

        self.__file_name = file_name

        self.__path_name = path_name

        self.channel_names = None

        self.dimensions = dimensions

        self.__spacing = spacing

        self.__channelstack = channelstack

    @property
    def multichannel(self):
        return True if self.pixel_data.ndim == self.dimensions + 1 else False

    @property
    def volumetric(self):
        if self.dimensions == 3:
            return True

        return False

    @property
    def spacing(self):
        if self.__spacing is not None:
            return tuple(numpy.divide(self.__spacing, self.__spacing[1]))

        if self.parent_image is None:
            return (1.0,) * self.dimensions

        return self.parent_image.spacing

    @spacing.setter
    def spacing(self, spacing):
        self.__spacing = spacing

    def get_image(self):
        """Return the primary image"""
        return self.__image

    def set_image(self, image, convert=True):
        """Set the primary image

        Convert the image to a numpy array of dtype = np.float64.
        Rescale according to Matlab's rules for im2double:
        * single/double values: keep the same
        * uint8/16/32/64: scale 0 to max to 0 to 1
        * int8/16/32/64: scale min to max to 0 to 1
        * logical: save as is (and get if must_be_binary)
        """
        img = numpy.asanyarray(image)
        if img.dtype.name == "bool" or not convert:
            self.__image = img
            return
        mval = 0.0
        scale = 1.0
        fix_range = False
        # unsigend and signed integer types
        if numpy.issubdtype(img.dtype, numpy.integer):
            scale = float(numpy.iinfo(img.dtype).max)
        # signed integer types need shifting
        if numpy.issubdtype(img.dtype, numpy.signedinteger):
            mval = float(numpy.iinfo(img.dtype).min)
            fix_range = True
        # Avoid temporaries by doing the shift/scale in place.
        img = img.astype(numpy.float32)
        img -= mval
        img /= scale
        if fix_range:
            # These types will always have ranges between 0 and 1. Make it so.
            numpy.clip(img, 0, 1, out=img)
        self.__image = img

    image = property(get_image, set_image)
    pixel_data = property(get_image, set_image)

    @property
    def has_parent_image(self):
        """True if this image has a defined parent"""
        return self.parent_image is not None

    @property
    def has_masking_objects(self):
        """True if the image was cropped with objects

        If this is true, there will also be a valid labels matrix
        available through the labels property
        """
        return self.masking_objects is not None

    @property
    def labels(self):
        """Get the segmentation labels from the masking objects

        returns the "segmented" labels: others are available through
        the masking_object.
        """
        if not self.has_masking_objects:
            return None

        return self.crop_image_similarly(self.masking_objects.segmented)

    @property
    def mask(self):
        """Return the mask (pixels to be considered) for the primary image
        """
        if self.__mask is not None:
            return self.__mask

        if self.has_masking_objects:
            return self.crop_image_similarly(self.crop_mask)

        if self.has_parent_image:
            mask = self.parent_image.mask

            return self.crop_image_similarly(mask)

        image = self.image

        #
        # Exclude channel, if present, from shape
        #
        shape = image.shape

        if self.multichannel:
            shape = shape[:-1]

        return numpy.ones(shape, dtype=bool)

    @mask.setter
    def mask(self, mask):
        """Set the mask (pixels to be considered) for the primary image

        Convert the input into a numpy array. If the input is numeric,
        we convert it to boolean by testing each element for non-zero.
        """
        m = numpy.array(mask)

        if not (m.dtype.type is bool):
            m = m != 0

        self.__mask = m
        self.__has_mask = True

    @property
    def has_mask(self):
        """True if the image has a mask"""
        if self.__has_mask:
            return True

        if self.has_crop_mask:
            return True

        if self.parent_image is not None:
            return self.parent_image.has_mask

        return False

    @property
    def crop_mask(self):
        """Return the mask used to crop this image"""
        if self.__crop_mask is not None:
            return self.__crop_mask

        if self.has_masking_objects:
            return self.masking_objects.segmented != 0

        if self.has_parent_image:
            return self.parent_image.crop_mask

        #
        # If no crop mask, return the mask which should be all ones
        #
        return self.mask

    @crop_mask.setter
    def crop_mask(self, crop_mask):
        self.__has_crop_mask = True
        self.__crop_mask = crop_mask

    @property
    def has_crop_mask(self):
        """True if the image or its ancestors has a crop mask"""
        return (
            self.__has_crop_mask
            or self.has_masking_objects
            or (self.has_parent_image and self.parent_image.has_crop_mask)
        )

    def crop_image_similarly(self, image):
        """Crop a 2-d or 3-d image using this image's crop mask

        image - a np.ndarray to be cropped (of any type)
        """
        if image.shape[:2] == self.pixel_data.shape[:2]:
            # Same size - no cropping needed
            return image
        if any(
            [
                my_size > other_size
                for my_size, other_size in zip(self.pixel_data.shape, image.shape)
            ]
        ):
            raise ValueError(
                "Image to be cropped is smaller: %s vs %s"
                % (repr(image.shape), repr(self.pixel_data.shape))
            )
        if not self.has_crop_mask:
            raise RuntimeError(
                "Images are of different size and no crop mask available.\n"
                "Use the Crop and Align modules to match images of different sizes."
            )
        cropped_image = crop_image(image, self.crop_mask)
        if cropped_image.shape[0:2] != self.pixel_data.shape[0:2]:
            raise ValueError(
                "Cropped image is not the same size as the reference image: %s vs %s"
                % (repr(cropped_image.shape), repr(self.pixel_data.shape))
            )
        return cropped_image

    @property
    def file_name(self):
        """The name of the file holding this image

        If the image is derived, then return the file name of the first
        ancestor that has a file name. Return None if the image does not have
        an ancestor or if no ancestor has a file name.
        """
        if self.__file_name is not None:
            return self.__file_name
        elif self.has_parent_image:
            return self.parent_image.file_name
        else:
            return None

    @property
    def path_name(self):
        """The path to the file holding this image

        If the image is derived, then return the path name of the first
        ancestor that has a path name. Return None if the image does not have
        an ancestor or if no ancestor has a file name.
        """
        if not self.__path_name is None:
            return self.__path_name
        elif self.has_parent_image:
            return self.parent_image.path_name
        else:
            return None

    @property
    def has_channel_names(self):
        """True if there are channel names on this image"""
        return self.channel_names is not None

    @property
    def scale(self):
        """The scale at acquisition

        This is the intensity scale used by the acquisition device. For
        instance, a microscope might use a 12-bit a/d converter to acquire
        an image and store that information using the TIF MaxSampleValue
        tag = 4095.
        """
        if self.__scale is None and self.has_parent_image:
            return self.parent_image.scale

        return self.__scale

    @property
    def channelstack(self):
        if self.__channelstack is not None:
            return self.__channelstack

    @channelstack.setter
    def channelstack(self, channelstack):
        self.__channelstack = channelstack



import numpy

from ..constants.image import CT_OBJECTS
from ..constants.measurement import COLTYPE_INTEGER, COLTYPE_VARCHAR, IMAGE
from ..constants.measurement import C_FILE_NAME
from ..constants.measurement import C_METADATA
from ..constants.measurement import C_OBJECTS_FILE_NAME
from ..constants.measurement import C_OBJECTS_PATH_NAME
from ..constants.measurement import C_PATH_NAME
from ..constants.measurement import EXPERIMENT
from ..constants.measurement import GROUP_INDEX
from ..constants.measurement import GROUP_NUMBER
from ..constants.measurement import GROUP_LENGTH
from ..constants.measurement import M_GROUPING_TAGS
from ..measurement import Measurements
from ..module import Module
from ..pipeline import Pipeline
from ..setting import Binary
from ..setting import Divider
from ..setting import HTMLText
from ..setting import HiddenCount
from ..setting import SettingsGroup
from ..setting import Table
from ..setting.choice import Choice
from ..setting.do_something import DoSomething
from ..setting.do_something import RemoveSettingButton
from ..utilities.image import image_resource

__doc__ = """\
Groups
======

The **Groups** module organizes sets of images into groups.

Once the images have been identified with the **Images** module, have
had metadata associated with them using the **Metadata** module, and
have been assigned names by the **NamesAndTypes** module, you have the
option of further sub-dividing the image sets into groups that share a
common feature. Some downstream modules of CellProfiler are capable of
processing groups of images in useful ways (e.g., object tracking within
a set of images comprising a time series, illumination correction within
a set of images comprising an experimental batch, data export for a set
of images comprising a plate).

|

============ ============ ===============
Supports 2D? Supports 3D? Respects masks?
============ ============ ===============
YES          YES          NO
============ ============ ===============

What is an image “group”?
^^^^^^^^^^^^^^^^^^^^^^^^^

The key to understanding why grouping may be necessary is that
CellProfiler processes the input images sequentially and in the order
given by the NamesAndTypes module. If you have multiple collections (or
“groups”) of images that should be processed independently from each
other, CellProfiler will simply finish processing one collection and
proceed to the next, ignoring any distinction between them unless told
otherwise via the **Groups** module.

To illustrate this idea, below are two examples where the grouping
concept can be useful or important:

-  If you have time-lapse movie data that is in the form of individual
   image files, and you are performing object tracking, it is important
   to indicate to CellProfiler that the end of a movie indicates the end
   of a distinct data set. Without doing so, CellProfiler will simply
   take the first frame of the next movie as a continuation of the
   previous one. If each set of files that comprise a movie is defined
   using the **Metadata** module, the relevant metadata can be used in
   this module to insure that object tracking only takes place within
   each movie.
-  If you are performing illumination correction for a screening
   experiment, we recommend that the illumination function (an image
   which represents the overall background fluorescence) be calculated
   on a per-plate basis. Since the illumination function is an aggregate
   of images from a plate, running a pipeline must yield a single
   illumination function for each plate. Running this pipeline multiple
   times, once for each plate, will give the desired result but would be
   tedious and time-consuming. In this case, CellProfiler can use image
   grouping for this purpose; if plate metadata can be defined by the
   **Metadata** module, grouping will enable you to process images that
   have the same plate metadata together.

What are the inputs?
^^^^^^^^^^^^^^^^^^^^

Using this module assumes that you have already adjusted the following
Input modules:

-  Used the **Images** module to produce a list of images to analyze.
-  Used the **Metadata** module to produce metadata defining the
   distinct sub-divisions between groups of images.
-  Used the **NamesAndTypes** module to assign names to individual
   channels and create image sets.

Selecting this module
will display a panel, allowing you to select whether you want to create
groups or not. A grouping may be defined as according to any or as many
of the metadata categories as defined by the **Metadata** module. By
selecting a metadata tag from the drop-down for the metadata category,
the **Groups** module will sub-divide and assemble the image sets
according to their unique metadata value. Upon adding a metadata
category, the two tables underneath will update to show the resultant
organization of the image sets for each group.

What do I get as output?
^^^^^^^^^^^^^^^^^^^^^^^^

The final product of the **Groups** module is a list defining subsets of
image sets that will be processed independently of the other subsets.

-  If no groups are defined, the Analysis modules in the rest of the
   pipeline will be applied to all images in exactly the same way.
-  If groups are defined in the **Groups** module, then organizationally
   (and transparently to you), CellProfiler will begin the analyses with
   the first image set of the group, end with the last image set of the
   group, and then proceed to the next group.

The two tables at the bottom provide the following information when a
metadata category is selected:

-  The *grouping list* (top table) shows the unique values of the
   selected metadata under the “Group” column; each of the unique values
   comprises a group. The “Count” column shows the number of image sets
   included in a given group; this is useful as a “sanity check” to make
   sure that the expected numbers of images are present.
-  The *image set list* (bottom table) shows the file name and location
   of each of the image sets that comprise the groups.


.. image:: {GROUPS_DISPLAY_TABLE}
   :width: 100%


Measurements made by this module
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

-  *Group\_Number:* The index of each grouping, as defined by the unique
   combinations of the metadata tags specified. These are written to the
   per-image table.
-  *Group\_Index:* The index of each image set within each grouping, as
   defined by the *Group\_Number*. These are written to the per-image
   table.

Technical notes
^^^^^^^^^^^^^^^

To perform grouping, only one analysis worker (i.e., copy of
CellProfiler) will be allocated to handle each group. This means that
you may have multiple workers created (as set under the Preferences),
but only a subset of them may actually be active, depending on the
number of groups you have.
""".format(
    **{"GROUPS_DISPLAY_TABLE": image_resource("Groups_ExampleDisplayTable.png")}
)


class Groups(Module):
    variable_revision_number = 2
    module_name = "Groups"
    category = "File Processing"

    IDX_GROUPING_METADATA_COUNT = 1

    def create_settings(self):
        self.pipeline = None
        self.metadata_keys = {}

        module_explanation = [
            "The %s module optionally allows you to split your list of images into image subsets"
            % self.module_name,
            "(groups) which will be processed independently of each other. Examples of",
            "groupings include screening batches, microtiter plates, time-lapse movies, etc.",
        ]
        self.set_notes([" ".join(module_explanation)])

        self.wants_groups = Binary(
            "Do you want to group your images?",
            False,
            doc="""\
Select "*{YES}*" if you need to split your images into image subsets (or
*groups*) such that each group is processed independently of each other.
See the main module help for more details.
""".format(
                **{"YES": "Yes"}
            ),
        )

        self.grouping_text = HTMLText(
            "",
            content="Each unique metadata value (or combination of values) will be defined as a group",
            size=(30, 2),
        )

        self.grouping_metadata = []

        self.grouping_metadata_count = HiddenCount(
            self.grouping_metadata, "grouping metadata count"
        )

        self.add_grouping_metadata(can_remove=False)

        self.add_grouping_metadata_button = DoSomething(
            "", "Add another metadata item", self.add_grouping_metadata
        )

        self.grouping_list = Table(
            "Grouping list",
            min_size=(300, 100),
            doc="""\
This list shows the unique values of the selected metadata under the
“Group” column; each of the unique values comprises a group. The “Count”
column shows the number of image sets that included in a given group;
this is useful as a “sanity check”, to make sure that the expected
number of images are present. For example, if you are grouping by
per-plate metadata from a 384-well assay with 2 sites per well
consisting of 3 plates, you would expect to see 3 groups (each from the
3 unique plate IDs), with 384 wells × 2 sites/well = 768 image sets in
each.
""",
        )

        self.image_set_list = Table(
            "Image sets",
            doc="""\
This list displays the file name and location of each of the image sets
that comprise the group. For example, if you are grouping by per-plate
metadata from a 384-well assay with 2 sites per well consisting of 3
plates, you would expect to see a table consisting of 3 plates × 384
wells/plate ×2 sites/well = 2304 rows.
""",
        )

    def add_grouping_metadata(self, can_remove=True):
        group = SettingsGroup()
        self.grouping_metadata.append(group)

        def get_group_metadata_choices(pipeline):
            choices = self.get_metadata_choices(pipeline, group)
            if len(choices) == 0:
                choices.append("None")
            return choices

        if self.pipeline is not None:
            choices = get_group_metadata_choices(self.pipeline)
        else:
            choices = ["None"]

        group.append(
            "metadata_choice",
            Choice(
                "Metadata category",
                choices,
                choices_fn=get_group_metadata_choices,
                doc="""\
Specify the metadata category with which to define a group. Once a
selection is made, the two listings below will display the updated
values:

-  The *grouping list* (top table) shows the unique values of the
   selected metadata under the “Group” column; each of the unique values
   comprises a group. The “Count” column shows the number of image sets
   included in a given group; this is useful as a “sanity check” to make
   sure that the expected numbers of images are present.
-  The *image set list* (bottom table) shows the file name and location
   of each of the image sets that comprise the groups. In this example,
   the table has 26 rows, one for each of the DNA and GFP image sets
   defined by the **NamesAndTypes** module.

You may specify multiple metadata tags to group with by clicking the
“Add” button. This would be necessary if a combination of metadata is
required in order to define a group. Upon adding a metadata category,
the two tables will update in the panels below showing the resulting
organization of the image data for each group.

As an example, an time-lapse experiment consists of a set of movie
images (indexed by a frame number), collected on a per-well basis. The
plate, well, wavelength and frame number metadata have been extracted
using the **Metadata** module. Using the **NamesAndTypes** module, the
two image channels (OrigBlue, *w1* and OrigGreen, *w2*) have been set up
in the following way:

+------------------------+-------------------------------+--------------------------------+-------------+------------+-------------------+
| **Image set number**   | **OrigBlue (w1) file name**   | **OrigGreen (w2) file name**   | **Plate**   | **Well**   | **FrameNumber**   |
+========================+===============================+================================+=============+============+===================+
| 1                      | P-12345\_A01\_t001\_w1.tif    | P-12345\_A01\_t001\_w2.tif     | P-12345     | A01        | t001              |
+------------------------+-------------------------------+--------------------------------+-------------+------------+-------------------+
| 2                      | P-12345\_A01\_t002\_w1.tif    | P-12345\_A01\_t002\_w2.tif     | P-12345     | A01        | t002              |
+------------------------+-------------------------------+--------------------------------+-------------+------------+-------------------+
| 3                      | P-12345\_B01\_t001\_w1.tif    | P-12345\_B01\_t001\_w2.tif     | P-12345     | B01        | t001              |
+------------------------+-------------------------------+--------------------------------+-------------+------------+-------------------+
| 4                      | P-12345\_B01\_t002\_w1.tif    | P-12345\_B01\_t002\_w2.tif     | P-12345     | B01        | t002              |
+------------------------+-------------------------------+--------------------------------+-------------+------------+-------------------+
| 5                      | 2-ABCDF\_A01\_t001\_w1.tif    | 2-ABCDF\_A01\_t001\_w2.tif     | 2-ABCDF\_   | A01        | t001              |
+------------------------+-------------------------------+--------------------------------+-------------+------------+-------------------+
| 6                      | 2-ABCDF\_A01\_t002\_w1.tif    | 2-ABCDF\_A01\_t002\_w2.tif     | 2-ABCDF\_   | A01        | t002              |
+------------------------+-------------------------------+--------------------------------+-------------+------------+-------------------+
| 7                      | 2-ABCDF\_B01\_t001\_w1.tif    | 2-ABCDF\_B01\_t001\_w2.tif     | 2-ABCDF\_   | B01        | t001              |
+------------------------+-------------------------------+--------------------------------+-------------+------------+-------------------+
| 8                      | 2-ABCDF\_B01\_t002\_w1.tif    | 2-ABCDF\_B01\_t002\_w2.tif     | 2-ABCDF\_   | B01        | t002              |
+------------------------+-------------------------------+--------------------------------+-------------+------------+-------------------+

We would like to perform object tracking for each movie, i.e., for each
plate and well. Without the use of groups, even though image sets 1–2,
3–4, 5–6, and 7–8 represent different movies, image set 3 will get
processed immediately after image set 2, image set 5 after 4, and so on.
For an object tracking assay, failure to recognize where the movies
start and end would lead to incorrect tracking results.

Selecting the *Plate* followed by the *Well* metadata as the metadata
categories will create four groups based on the unique plate and well
combinations:

+---------------------+----------------------+------------------------+-------------+------------+------------------------------+------------------------------+------------------------------+
| **Grouping tags**                          | **Image set tags**                                                               | **Channels**                                                |
+=====================+======================+========================+=============+============+==============================+==============================+==============================+
| **Group number**    | **Group index**      | **Image set number**   | **Plate**   | **Well**   | **FrameNumber**              | **OrigBlue**                 | **OrigGreen**                |
+---------------------+----------------------+------------------------+-------------+------------+------------------------------+------------------------------+------------------------------+
| 1                   | 1                    | 1                      | P-12345     | A01        | t001                         | P-12345\_A01\_t001\_w1.tif   | P-12345\_A01\_t001\_w2.tif   |
+---------------------+----------------------+------------------------+-------------+------------+------------------------------+------------------------------+------------------------------+
| 1                   | 2                    | 2                      | P-12345     | A01        | t002                         | P-12345\_A01\_t002\_w1.tif   | P-12345\_A01\_t002\_w2.tif   |
+---------------------+----------------------+------------------------+-------------+------------+------------------------------+------------------------------+------------------------------+
| 2                   | 1                    | 3                      | P-12345     | B01        | t001                         | P-12345\_B01\_t001\_w1.tif   | P-12345\_B01\_t001\_w2.tif   |
+---------------------+----------------------+------------------------+-------------+------------+------------------------------+------------------------------+------------------------------+
| 2                   | 2                    | 4                      | P-12345     | B01        | t002                         | P-12345\_B01\_t002\_w1.tif   | P-12345\_B01\_t002\_w2.tif   |
+---------------------+----------------------+------------------------+-------------+------------+------------------------------+------------------------------+------------------------------+
| 3                   | 1                    | 5                      | 2-ABCDF     | A01        | t001                         | 2-ABCDF\_A01\_t001\_w1.tif   | 2-ABCDF\_A01\_t001\_w2.tif   |
+---------------------+----------------------+------------------------+-------------+------------+------------------------------+------------------------------+------------------------------+
| 3                   | 2                    | 6                      | 2-ABCDF     | A01        | t002                         | 2-ABCDF\_A01\_t002\_w1.tif   | 2-ABCDF\_A01\_t002\_w2.tif   |
+---------------------+----------------------+------------------------+-------------+------------+------------------------------+------------------------------+------------------------------+
| 4                   | 1                    | 7                      | 2-ABCDF     | B01        | t001                         | 2-ABCDF\_B01\_t001\_w1.tif   | 2-ABCDF\_B01\_t001\_w2.tif   |
+---------------------+----------------------+------------------------+-------------+------------+------------------------------+------------------------------+------------------------------+
| 4                   | 2                    | 8                      | 2-ABCDF     | B01        | t002                         | 2-ABCDF\_B01\_t002\_w1.tif   | 2-ABCDF\_B01\_t002\_w2.tif   |
+---------------------+----------------------+------------------------+-------------+------------+------------------------------+------------------------------+------------------------------+

Each group will be processed independently from the others, which is the
desired behavior.
""",
            ),
        )

        group.append("divider", Divider())

        group.can_remove = can_remove

        if can_remove:
            group.append(
                "remover",
                RemoveSettingButton(
                    "", "Remove this metadata item", self.grouping_metadata, group
                ),
            )

    def get_metadata_choices(self, pipeline, group):
        if self.pipeline is not None:
            return sorted(self.metadata_keys)
        #
        # Unfortunate - an expensive operation to find the possible metadata
        #               keys from one of the columns in an image set.
        # Just fake it into having something that will work
        #
        return [group.metadata_choice.value]

    def settings(self):
        result = [self.wants_groups, self.grouping_metadata_count]
        for group in self.grouping_metadata:
            result += [group.metadata_choice]
        return result

    def visible_settings(self):
        result = [self.wants_groups]
        if self.wants_groups:
            for group in self.grouping_metadata:
                result += [group.metadata_choice]
                if group.can_remove:
                    result += [group.remover]
                result += [group.divider]
            result += [
                self.add_grouping_metadata_button,
                self.grouping_list,
                self.image_set_list,
            ]
        return result

    def help_settings(self):
        result = [
            self.wants_groups,
            self.grouping_metadata[0].metadata_choice,
            self.grouping_list,
            self.image_set_list,
        ]
        return result

    def prepare_settings(self, setting_values):
        nmetadata = int(setting_values[self.IDX_GROUPING_METADATA_COUNT])
        while len(self.grouping_metadata) > nmetadata:
            del self.grouping_metadata[-1]

        while len(self.grouping_metadata) < nmetadata:
            self.add_grouping_metadata()

    def on_activated(self, workspace):
        self.pipeline = workspace.pipeline
        self.workspace = workspace
        assert isinstance(self.pipeline, Pipeline)
        if self.wants_groups:
            self.metadata_keys = []
            self.image_sets_initialized = workspace.refresh_image_set()
            self.metadata_keys = list(
                self.pipeline.get_available_metadata_keys().keys()
            )
            is_valid = True
            for group in self.grouping_metadata:
                try:
                    group.metadata_choice.test_valid(self.pipeline)
                except:
                    is_valid = False
            if is_valid:
                self.update_tables()
        else:
            self.image_sets_initialized = False

    def on_deactivated(self):
        self.pipeline = None

    def on_setting_changed(self, setting, pipeline):
        if (
            setting == self.wants_groups
            and self.wants_groups
            and not self.image_sets_initialized
        ):
            workspace = self.workspace
            self.on_deactivated()
            self.on_activated(workspace)
            needs_prepare_run = False
        else:
            needs_prepare_run = True
        #
        # Unfortunately, test_valid has the side effect of getting
        # the choices set which is why it's called here
        #
        is_valid = True
        for group in self.grouping_metadata:
            try:
                group.metadata_choice.test_valid(pipeline)
            except:
                is_valid = False
        if is_valid:
            if needs_prepare_run:
                result = self.prepare_run(self.workspace, changed_setting = True)
                if not result:
                    return
            self.update_tables()

    def update_tables(self):
        if self.wants_groups:
            try:
                if not self.workspace.refresh_image_set():
                    return
            except:
                return
            m = self.workspace.measurements
            assert isinstance(m, Measurements)
            channel_descriptors = m.get_channel_descriptors()

            self.grouping_list.clear_columns()
            self.grouping_list.clear_rows()
            self.image_set_list.clear_columns()
            self.image_set_list.clear_rows()
            metadata_key_names = [
                group.metadata_choice.value
                for group in self.grouping_metadata
                if group.metadata_choice.value != "None"
            ]
            metadata_feature_names = [
                "_".join((C_METADATA, key)) for key in metadata_key_names
            ]
            metadata_key_names = [
                x[(len(C_METADATA) + 1) :] for x in metadata_feature_names
            ]
            image_set_feature_names = [
                GROUP_NUMBER,
                GROUP_INDEX,
            ] + metadata_feature_names
            self.image_set_list.insert_column(0, "Group number")
            self.image_set_list.insert_column(1, "Group index")

            for i, key in enumerate(metadata_key_names):
                for l, offset in ((self.grouping_list, 0), (self.image_set_list, 2)):
                    l.insert_column(i + offset, "Group: %s" % key)

            self.grouping_list.insert_column(len(metadata_key_names), "Count")

            image_numbers = m.get_image_numbers()
            group_indexes = m["Image", GROUP_INDEX, image_numbers,][:]
            group_numbers = m["Image", GROUP_NUMBER, image_numbers,][:]
            counts = numpy.bincount(group_numbers)
            first_indexes = numpy.argwhere(group_indexes == 1).flatten()
            group_keys = [
                m["Image", feature, image_numbers] for feature in metadata_feature_names
            ]
            k_count = sorted(
                [
                    (
                        group_numbers[i],
                        [x[i] for x in group_keys],
                        counts[group_numbers[i]],
                    )
                    for i in first_indexes
                ]
            )
            for group_number, group_key_values, c in k_count:
                row = group_key_values + [c]
                self.grouping_list.data.append(row)

            for image_name, channel_type in channel_descriptors.items():
                idx = len(image_set_feature_names)
                self.image_set_list.insert_column(idx, "Path: %s" % image_name)
                self.image_set_list.insert_column(idx + 1, "File: %s" % image_name)
                if channel_type == CT_OBJECTS:
                    image_set_feature_names.append(
                        C_OBJECTS_PATH_NAME + "_" + image_name
                    )
                    image_set_feature_names.append(
                        C_OBJECTS_FILE_NAME + "_" + image_name
                    )
                else:
                    image_set_feature_names.append(C_PATH_NAME + "_" + image_name)
                    image_set_feature_names.append(C_FILE_NAME + "_" + image_name)

            all_features = [
                m["Image", ftr, image_numbers] for ftr in image_set_feature_names
            ]
            order = numpy.lexsort((group_indexes, group_numbers))

            for idx in order:
                row = [str(x[idx]) for x in all_features]
                self.image_set_list.data.append(row)

    def get_groupings(self, workspace):
        """Return the image groupings of the image sets in an image set list

        returns a tuple of key_names and group_list:
        key_names - the names of the keys that identify the groupings
        group_list - a sequence composed of two-tuples.
                     the first element of the tuple has the values for
                     the key_names for this group.
                     the second element of the tuple is a sequence of
                     image numbers comprising the image sets of the group
        For instance, an experiment might have key_names of 'Metadata_Row'
        and 'Metadata_Column' and a group_list of:
        [ ({'Row':'A','Column':'01'), [0,96,192]),
          (('Row':'A','Column':'02'), [1,97,193]),... ]
        """
        if not self.wants_groups:
            return
        key_list = self.get_grouping_tags_or_metadata()
        m = workspace.measurements
        for key in key_list:
            if key not in m.get_feature_names("Image"):
                if key.startswith(C_METADATA):
                    key = key[len(C_METADATA) + 1 :]
                workspace.pipeline.report_prepare_run_error(
                    self,
                    (
                        'The groups module is misconfigured. "%s" was chosen as\n'
                        "one of the metadata tags, but that metadata tag is not\n"
                        "defined in the Metadata module."
                    )
                    % key,
                )
                return None
        return key_list, m.get_groupings(key_list)

    def get_grouping_tags_or_metadata(self):
        """Return the metadata keys used for grouping"""
        if not self.wants_groups:
            return None
        return [
            "_".join((C_METADATA, g.metadata_choice.value,))
            for g in self.grouping_metadata
        ]

    def change_causes_prepare_run(self, setting):
        """Return True if changing the setting passed changes the image sets

        setting - the setting that was changed
        """
        return setting in self.settings()

    def is_load_module(self):
        """Marks this module as a module that affects the image sets

        Groups is a load module because it can reorder image sets, but only
        if grouping is turned on.
        """
        return self.wants_groups.value

    @classmethod
    def is_input_module(self):
        return True

    def prepare_run(self, workspace, changed_setting = False):
        """Reorder the image sets and assign group number and index"""
        if workspace.pipeline.in_batch_mode():
            return True

        if not self.wants_groups:
            return True

        for group in self.grouping_metadata:
            if group.metadata_choice.value == "None":
                return False

        if len(workspace.measurements.get_image_numbers()) == 0:
            # Refresh image set to make sure it's actually empty, if and only if there was a settings change
            if changed_setting:
                workspace.refresh_image_set()
                if len(workspace.measurements.get_image_numbers()) == 0:
                    return False
            else:
                return False

        result = self.get_groupings(workspace)
        if result is None:
            return False
        key_list, groupings = result
        #
        # Create arrays of group number, group_index and image_number
        #
        group_numbers = numpy.hstack(
            [
                numpy.ones(len(image_numbers), int) * (i + 1)
                for i, (keys, image_numbers) in enumerate(groupings)
            ]
        )
        group_indexes = numpy.hstack(
            [numpy.arange(len(image_numbers)) + 1 for keys, image_numbers in groupings]
        )
        group_lens = numpy.hstack(
            [
                numpy.ones(len(image_numbers), int) * (len(image_numbers))
                for i, (keys, image_numbers) in enumerate(groupings)
            ]
        )
        image_numbers = numpy.hstack(
            [image_numbers for keys, image_numbers in groupings]
        )
        order = numpy.lexsort((group_indexes, group_numbers))
        group_numbers = group_numbers[order]
        group_indexes = group_indexes[order]
        group_lens = group_lens[order]

        m = workspace.measurements
        assert isinstance(m, Measurements)
        #
        # Downstream processing requires that image sets be ordered by
        # increasing group number, then increasing group index.
        #
        new_image_numbers = numpy.zeros(numpy.max(image_numbers) + 1, int)
        new_image_numbers[image_numbers[order]] = numpy.arange(len(image_numbers)) + 1
        m.reorder_image_measurements(new_image_numbers)
        m.add_all_measurements(
            "Image", GROUP_NUMBER, group_numbers,
        )
        m.add_all_measurements(
            "Image", GROUP_INDEX, group_indexes,
        )
        m.add_all_measurements(
            "Image", GROUP_LENGTH, group_lens,
        )
        m.set_grouping_tags(self.get_grouping_tags_or_metadata())
        return True

    def run(self, workspace):
        pass

    def get_measurement_columns(self, pipeline):
        """Return the measurements recorded by this module

        GroupNumber and GroupIndex are accounted for by the pipeline itself.
        """
        result = []
        if self.wants_groups:
            result.append((EXPERIMENT, M_GROUPING_TAGS, COLTYPE_VARCHAR,))
            result.append((IMAGE, GROUP_LENGTH, COLTYPE_INTEGER))
            #
            # These are bound to be produced elsewhere, but it is quite
            # computationally expensive to find that out. If they are
            # duplicated by another module, no big deal.
            #
            for ftr in self.get_grouping_tags_or_metadata():
                result.append(("Image", ftr, COLTYPE_VARCHAR,))
        return result

    def upgrade_settings(self, setting_values, variable_revision_number, module_name):
        if variable_revision_number == 1:
            #
            # Remove the image name from the settings
            #
            new_setting_values = setting_values[
                : (self.IDX_GROUPING_METADATA_COUNT + 1)
            ]
            for i in range(int(setting_values[self.IDX_GROUPING_METADATA_COUNT])):
                new_setting_values.append(
                    setting_values[self.IDX_GROUPING_METADATA_COUNT + 2 + i * 2]
                )
            setting_values = new_setting_values
            variable_revision_number = 2
        return setting_values, variable_revision_number

    def volumetric(self):
        return True



import re

from ._setting import Setting
from ._validation_error import ValidationError


class RegexpText(Setting):
    """A setting with a regexp button on the side
    """

    GUESS_FILE = "file"
    GUESS_FOLDER = "folder"

    def __init__(self, text, value, *args, **kwargs):
        """initialize the setting

        text   - the explanatory text for the setting
        value  - the default or initial value for the setting
        doc - documentation for the setting
        get_example_fn - a function that returns an example string for the
                         metadata editor
        guess - either GUESS_FILE to use potential file-name regular expressions
                when guessing in the regexp editor or GUESS_FOLDER to
                use folder-name guesses.
        """
        kwargs = kwargs.copy()
        self.get_example_fn = kwargs.pop("get_example_fn", None)
        self.guess = kwargs.pop("guess", self.GUESS_FILE)
        super(RegexpText, self).__init__(text, value, *args, **kwargs)

    def test_valid(self, pipeline):
        try:
            # Convert Matlab to Python
            pattern = re.sub("(\\(\\?)([<][^)>]+?[>])", "\\1P\\2", self.value)
            re.search("(|(%s))" % pattern, "")
        except re.error as v:
            raise ValidationError("Invalid regexp: %s" % v, self)



from ._module import Module
from ..image import Image
from ..setting.subscriber import ImageSubscriber
from ..setting.text import ImageName


class ImageProcessing(Module):
    category = "Image Processing"

    def create_settings(self):
        self.x_name = ImageSubscriber(
            "Select the input image", doc="Select the image you want to use."
        )

        self.y_name = ImageName(
            "Name the output image",
            self.__class__.__name__,
            doc="Enter the name you want to call the image produced by this module.",
        )

    def display(self, workspace, figure, cmap=None):
        if cmap is None:
            cmap = ["gray", "gray"]
        layout = (2, 1)

        figure.set_subplots(
            dimensions=workspace.display_data.dimensions, subplots=layout
        )

        figure.subplot_imshow(
            colormap=cmap[0],
            image=workspace.display_data.x_data,
            title=self.x_name.value,
            x=0,
            y=0,
        )

        figure.subplot_imshow(
            colormap=cmap[1],
            image=workspace.display_data.y_data,
            sharexy=figure.subplot(0, 0),
            title=self.y_name.value,
            x=1,
            y=0,
        )

    def run(self, workspace):
        x_name = self.x_name.value

        y_name = self.y_name.value

        images = workspace.image_set

        x = images.get_image(x_name)

        dimensions = x.dimensions

        x_data = x.pixel_data

        args = (setting.value for setting in self.settings()[2:])

        y_data = self.function(x_data, *args)

        y = Image(dimensions=dimensions, image=y_data, parent_image=x, convert=False)

        images.add(y_name, y)

        if self.show_window:
            workspace.display_data.x_data = x_data

            workspace.display_data.y_data = y_data

            workspace.display_data.dimensions = dimensions

    def settings(self):
        return [self.x_name, self.y_name]

    def visible_settings(self):
        return [self.x_name, self.y_name]

    def volumetric(self):
        return True
